#include <stdlib.h>

#ifdef WIN32
#else
#include <unistd.h>
#include <dirent.h>
#include "cam.h"
#include "pmt-proxy.h"
#endif

#include <signal.h>
#include <sys/types.h>

#include "common.h"
#include "box.h"
#include "btkcam.h"
#include "console-server.h"
#include "btkcam-client.h"
#include "stats.h"

// ------------------------------------------- external declarations ------------------------------
extern btkcam_client *btk[10];
extern char	Blines[10][255];
extern int 	Bcnt;
// ------------------------------------------------------------------------------------------------
// -- extern declarations (misc)
extern statistics *stats;
extern int config_log_to_file;
extern int config_background_execution;
extern int config_log_to_console;

// -- extern declarations --console broadcast server related
extern int config_console_logs_broadcast;
extern int config_console_logs_broadcast_port;
extern console_server *remote_console_server;


int instance_detection_socket = 0;

int main(int argc, char *argv[])
{
	// -------- wait for another running instance to exit, if it doesn't, exit this process. --------
	if( check_running_instance() )
	{
		printf("ERROR: There is another instance of btkcam running already, exiting.\n");
		return 0;
	}
	
	// -- set signal handlers
#ifndef WIN32
	set_signal_handlers();
#endif
	

	// -------------------- instantiate statistics class and set started time -----------------------
	stats = new statistics();
	strcpy(stats->started_on, get_time());
	
	// ----------------------------------- misc initialization --------------------------------------
	display_help();
	
	if( config_read_settings() ) return 0;
	
#if (not defined(__X86__)) && (not defined(WIN32)) //writing logs in Linux X86 or WIN32 versions is unnecessary.
	if( config_log_to_file ) 
		_slog(NO_TRANS, _DEBUG_BASIC, 0, _logmf("************** Starting log on %s **************", get_time() ), 2);
#endif


	// -- process command line parameters
	if( process_command_line(argv, argc) ) return 2;
	
	// -- show configuration settings
	config_show_settings( NULL );
	
	// ----------------------------- show on-screen start-up message --------------------------------
	char message[1024];
	char line[1024];
	sprintf(line, _TITLE, _CAMD_NAME, _VERSION); 
	strcpy(message, line); 
	strcat(message, "\n");
	config_show_settings( line );
	strcat(message, line);

	// -- send process to background, if enabled
	if( config_background_execution ) if( send_to_background() < 0 ) return 3;
#ifndef WIN32
	// ---------------------------------- initialize cam daemon -------------------------------------
	//_slog(NO_TRANS, _DEBUG_BASIC, 0, "Initializing cam daemon...", 1);
	if( service::initialize() ) return 3;

#endif
	
	// ----------------------- start console logs broadcast server, if enabled ----------------------
	if( config_console_logs_broadcast )
		remote_console_server = new console_server(config_console_logs_broadcast_port);

	// ----------------------- start Btkcam Clients, if enabled -------------------------------------
	int ixx;
	if (Bcnt>0) for (ixx=0;ixx<Bcnt;ixx++)
		{
		if (btk[ixx] = new btkcam_client(Blines[ixx],10), !btk[ixx])
			{
			perror("BTKCAM client Failed");
			return(0);
			}
		btk[ixx]->dbg= config_log_to_console;
		}

	// -------------------- if boxtype requires a pmt-proxy, start the service. ---------------------
#ifndef WIN32
	switch( box::type )
	{
		case 2: //dgstation
			pmt_proxy::start(DGSTATION_DEVICES, box::get_cam_socket_path() );
			break;
			
		default:
			break;
	}
#endif

	
	// -- if this is compiled for Linux X86 or WIN32, then it will only work as offline debugger.
#if defined(__X86__) || defined(WIN32)
	_slog(NO_TRANS, _DEBUG_BASIC, 1, "", 1);
	_slog(NO_TRANS, _DEBUG_BASIC, 1, "WARNING: The Linux X86 and WIN32 versions are for offline debugging only.\n", 1);
#endif
		
	// ------------------------------------- start daemon loop --------------------------------------
	_slog(NO_TRANS, _DEBUG_BASIC, 1, "cam[di:-]: Ready.", 1);
	
	// -- the cam daemon only starts for versions that run on boxes.
#if defined(__POWERPC__) || defined(__MIPSEL__)
	unsigned int service_inactivity_count = 0;
	service* service_ptr;
	while( 1 )
	{
		if( service::poll_services() > 0 )
		{
			//process service messages/disconnects
			service_ptr = service::service_root;
			while( service_ptr )
			{
				switch( service_ptr->status )
				{
					case SLEEPING:
						service_ptr = service_ptr->next;
						break;
						
					case DATA_AVAILABLE:
						if( service_ptr->process_message() )
							service_ptr->status = DISCONNECTED;
						else 
							service_ptr = service_ptr->next;
						break;
						
					case DISCONNECTED:
						_slog(NO_TRANS, _DEBUG_BASIC, 1, 
							_logmf("cam[di:%d]: Client disconnected.", service_ptr->descramble_index), 1);
						if( service_ptr->next == NULL )
						{
							service::remove( service_ptr );
							service_ptr = NULL;
						}
						else
						{
							service_ptr = service_ptr->next;
							service::remove( service_ptr->previous );
						}
						break;
				}
			}
			
			//accept connection and create new service
			if( service::listener_status == DATA_AVAILABLE )
				if( !service::accept_service_request() ) break; //if accept fails, end process.
		}
		service::remove_inactive();

		//------------------------------------------------------------------------------------
		if( service::count() == 0 )
		{
			//after 1 minute of inactivity, disconnect any active card-servers.
			if( (service_inactivity_count++) == 240 )
			{
			_slog(NO_TRANS, _DEBUG_BASIC, 1, "Timeout. Checking B: lines...", 1);
			for (ixx=0;ixx<Bcnt;ixx++) 
				{
				btk[ixx]->checknet((unsigned char *)"ASK:AIV\0",7); 
				if (btk[ixx]->is_ready()) continue;
				delete(btk[ixx]);
				if (btk[ixx] = new btkcam_client(Blines[ixx],10), !btk[ixx])
					{
					perror("BTKCAM client Failed");
					return(0);
					}
				btk[ixx]->dbg= config_log_to_console;
				}	
			service_inactivity_count=0;
			}
		}
		//------------------------------------------------------------------------------------
	}
#else
	while( true ) //Linux X86 or WIN32 version, stay idle indefinitely.
	{
		usleep( 250000 );
	}
#endif
	
	prepare_exit();
	return 0;
}

void display_help()
{
	printf(_TITLE, _CAMD_NAME, _VERSION);
}

int process_command_line(char *argv[], int argc)
{
	for(int i=1;i<argc;i++)
	{
		// ----------------------------------------- show help --------------------------------------------
		if( !strcmp(argv[i], "-help" ) )
		{
			return 1;
		}
		
		// ---------------------------------------- set debug level ---------------------------------------
		if( !strcmp(argv[i], "-debug-level" ) )
		{
			if( (i + 1) <= (argc - 1) ) set_debug_level( atoi( argv[i + 1] ) );
			continue;
		}

		
	}
	return 0;
}


int send_to_background( void )
{
	int i = fork();
	if(i < 0)
	{
		_slog(NO_TRANS, _DEBUG_BASIC, 0, "ERROR: Could not send process to background. Terminating.",1);
		return -1;
	} 
	if(i > 0) { exit(0); }
	
	setsid(); // obtain a new process group (disconnects from console)
	return 0;
}

void set_signal_handlers( void )
{
	signal(SIGINT, catch_interrupt);	// --- set int (ctrl-c) signal handler
	signal(SIGTERM, catch_terminate);	// --- set term (kill) signal handler
}

void catch_interrupt(int sig_num)
{
	prepare_exit();
}

void catch_terminate(int sig_num)
{
	prepare_exit();
}

bool check_running_instance( void )
{
	int instance_detection_socket, reuse, tries;
	struct sockaddr_in listener_endpoint;
	
	instance_detection_socket = socket(PF_INET, SOCK_STREAM, 0);
	listener_endpoint.sin_family = AF_INET;
	listener_endpoint.sin_port = _htons(664);
	listener_endpoint.sin_addr.s_addr = _htonl(INADDR_ANY);
	setsockopt(instance_detection_socket, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse) );
	
	tries = 0;
	while( bind(instance_detection_socket, (struct sockaddr *)&listener_endpoint, sizeof(listener_endpoint)) < 0 )
	{
		tries++;
		if( tries == 10 ) return true;
		usleep(100000);
	}
	
	return false;
}

void prepare_exit( void )
{
	//avoid re-entry
	static bool reentry = false;
	if( reentry ) return; reentry = true;
	
	printf("btkcam: INT/TERM signal received. Shutting down.\n");
	
	if( remote_console_server ) { delete remote_console_server; remote_console_server = 0; }
	if( pmt_proxy::ready ) pmt_proxy::stop();
	
	service::finalize();	

	fflush(stdout);
	
	if( instance_detection_socket) close( instance_detection_socket );
	exit(0);
}
